from os import environ

import pytest

from tests.utils import check


@pytest.mark.skipif(environ.get("CI") != "true", reason="not in CI")
def test_shutdown(client):
    client.app.shutdown()
    with pytest.raises(AssertionError, match="qBittorrent crashed..."):
        check(lambda: client.app_version(), "")



import sys

import pytest

from qbittorrentapi import APINames
from qbittorrentapi.transfer import TransferInfoDictionary


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Transfer
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


def test_info(client):
    info = client.transfer_info()
    assert isinstance(info, TransferInfoDictionary)
    assert "connection_status" in info
    info = client.transfer.info
    assert isinstance(info, TransferInfoDictionary)
    assert "connection_status" in info


def test_speed_limits_mode(client):
    assert client.transfer_speed_limits_mode() in {"0", "1"}
    assert client.transfer.speed_limits_mode in {"0", "1"}

    original_mode = client.transfer.speed_limits_mode
    client.transfer_set_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode
    original_mode = client.transfer.speed_limits_mode
    client.transfer.set_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode

    original_mode = client.transfer.speed_limits_mode
    client.transfer_toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode
    client.transfer_toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode == original_mode
    original_mode = client.transfer.speed_limits_mode
    client.transfer.toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode != original_mode
    client.transfer.toggle_speed_limits_mode()
    assert client.transfer.speed_limits_mode == original_mode

    client.transfer_set_speed_limits_mode(intended_state=True)
    assert client.transfer.speed_limits_mode != "0"
    client.transfer_setSpeedLimitsMode(intended_state=False)
    assert client.transfer.speed_limits_mode != "1"

    client.transfer_toggle_speed_limits_mode(intended_state=True)
    assert client.transfer.speed_limits_mode != "0"
    client.transfer_toggle_speed_limits_mode(intended_state=False)
    assert client.transfer.speed_limits_mode != "1"

    client.transfer.speed_limits_mode = True
    assert client.transfer.speed_limits_mode != "0"
    client.transfer.speed_limits_mode = False
    assert client.transfer.speed_limits_mode != "1"

    client.transfer.speedLimitsMode = True
    assert client.transfer.speedLimitsMode != "0"
    client.transfer.speedLimitsMode = False
    assert client.transfer.speedLimitsMode != "1"


def test_download_limit(client):
    client.transfer_set_download_limit(limit=2048)
    assert client.transfer_download_limit() == 2048
    client.transfer_setDownloadLimit(limit=3072)
    assert client.transfer_downloadLimit() == 3072

    client.transfer.download_limit = 4096
    assert client.transfer.download_limit == 4096
    client.transfer.downloadLimit = 5120
    assert client.transfer.downloadLimit == 5120


def test_upload_limit(client):
    client.transfer_set_upload_limit(limit=2048)
    assert client.transfer_upload_limit() == 2048
    client.transfer_setUploadLimit(limit=3072)
    assert client.transfer_uploadLimit() == 3072

    client.transfer.upload_limit = 4096
    assert client.transfer.upload_limit == 4096
    client.transfer.uploadLimit = 5120
    assert client.transfer.uploadLimit == 5120


@pytest.mark.skipif_before_api_version("2.3")
def test_ban_peers(client):
    client.transfer_ban_peers(peers="1.1.1.1:8080")
    assert "1.1.1.1" in client.app.preferences.banned_IPs
    client.transfer.ban_peers(peers="1.1.1.2:8080")
    assert "1.1.1.2" in client.app.preferences.banned_IPs

    client.transfer_ban_peers(peers=["1.1.1.3:8080", "1.1.1.4:8080"])
    assert "1.1.1.3" in client.app.preferences.banned_IPs
    assert "1.1.1.4" in client.app.preferences.banned_IPs
    client.transfer.ban_peers(peers=["1.1.1.5:8080", "1.1.1.6:8080"])
    assert "1.1.1.5" in client.app.preferences.banned_IPs
    assert "1.1.1.6" in client.app.preferences.banned_IPs


@pytest.mark.skipif_after_api_version("2.3")
def test_ban_peers_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.transfer_ban_peers(peers="1.1.1.1:8080")



from __future__ import annotations

from functools import wraps
from typing import Iterable
from typing import Mapping
from typing import cast

from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.definitions import APIKwargsT
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.definitions import JsonValueT
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import ListInputT


class SearchResultsDictionary(Dictionary[JsonValueT]):
    """
    Response for :meth:`~SearchAPIMixIn.search_results`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-results>`_
    """  # noqa: E501


class SearchStatus(ListEntry):
    """Item in :class:`SearchStatusesList`"""


class SearchStatusesList(List[SearchStatus]):
    """
    Response for :meth:`~SearchAPIMixIn.search_status`

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-status>`_
    """  # noqa: E501

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchStatus, client=client)


class SearchCategory(ListEntry):
    """Item in :class:`SearchCategoriesList`"""


class SearchCategoriesList(List[SearchCategory]):
    """Response for :meth:`~SearchAPIMixIn.search_categories`"""

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchCategory, client=client)


class SearchPlugin(ListEntry):
    """Item in :class:`SearchPluginsList`"""


class SearchPluginsList(List[SearchPlugin]):
    """
    Response for :meth:`~SearchAPIMixIn.search_plugins`.

    Definition: `<https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#user-content-get-search-plugins>`_
    """

    def __init__(self, list_entries: ListInputT, client: SearchAPIMixIn | None = None):
        super().__init__(list_entries, entry_class=SearchPlugin, client=client)


class SearchAPIMixIn(AppAPIMixIn):
    """
    Implementation for all ``Search`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> search_job = client.search_start(pattern="Ubuntu", plugins="all", category="all")
        >>> client.search_stop(search_id=search_job.id)
        >>> # or
        >>> search_job.stop()
    """

    @property
    def search(self) -> Search:
        """
        Allows for transparent interaction with ``Search`` endpoints.

        See Search class for usage.
        """
        if self._search is None:
            self._search = Search(client=self)
        return self._search

    def search_start(
        self,
        pattern: str | None = None,
        plugins: Iterable[str] | None = None,
        category: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchJobDictionary:
        """
        Start a search. Python must be installed. Host may limit number of concurrent
        searches.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises Conflict409Error:

        :param pattern: term to search for
        :param plugins: list of plugins to use for searching (supports 'all' and
            'enabled')
        :param category: categories to limit search; dependent on plugins. (supports
            'all')
        """
        data = {
            "pattern": pattern,
            "plugins": self._list2string(plugins, "|"),
            "category": category,
        }
        return self._post_cast(
            _name=APINames.Search,
            _method="start",
            data=data,
            response_class=SearchJobDictionary,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_stop(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Stop a running search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :param search_id: ID of search job to stop
        """
        data = {"id": search_id}
        self._post(
            _name=APINames.Search,
            _method="stop",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_status(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchStatusesList:
        """
        Retrieve status of one or all searches.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:

        :param search_id: ID of search to get status; leave empty for status of all jobs
        """
        params = {"id": search_id}
        return self._get_cast(
            _name=APINames.Search,
            _method="status",
            params=params,
            response_class=SearchStatusesList,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_results(
        self,
        search_id: str | int | None = None,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        """
        Retrieve the results for the search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :raises Conflict409Error:

        :param search_id: ID of search job
        :param limit: number of results to return
        :param offset: where to start returning results
        """
        data = {"id": search_id, "limit": limit, "offset": offset}
        return self._post_cast(
            _name=APINames.Search,
            _method="results",
            data=data,
            response_class=SearchResultsDictionary,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_delete(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Delete a search job.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :raises NotFound404Error:
        :param search_id: ID of search to delete
        """
        data = {"id": search_id}
        self._post(
            _name=APINames.Search,
            _method="delete",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_categories(
        self,
        plugin_name: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchCategoriesList:
        """
        Retrieve categories for search.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1) and removed
        with qBittorrent v4.3.0 (Web API v2.6).

        :param plugin_name: Limit categories returned by plugin(s) (supports ``all`` and ``enabled``)
        """
        data = {"pluginName": plugin_name}
        return self._post_cast(
            _name=APINames.Search,
            _method="categories",
            data=data,
            response_class=SearchCategoriesList,
            version_introduced="2.1.1",
            version_removed="2.6",
            **kwargs,
        )

    def search_plugins(self, **kwargs: APIKwargsT) -> SearchPluginsList:
        """
        Retrieve details of search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).
        """
        return self._get_cast(
            _name=APINames.Search,
            _method="plugins",
            response_class=SearchPluginsList,
            version_introduced="2.1.1",
            **kwargs,
        )

    def search_install_plugin(
        self,
        sources: Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Install search plugins from either URL or file.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param sources: list of URLs or filepaths
        """
        data = {"sources": self._list2string(sources, "|")}
        self._post(
            _name=APINames.Search,
            _method="installPlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_installPlugin = search_install_plugin

    def search_uninstall_plugin(
        self,
        names: Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Uninstall search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param names: names of plugins to uninstall
        """
        data = {"names": self._list2string(names, "|")}
        self._post(
            _name=APINames.Search,
            _method="uninstallPlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_uninstallPlugin = search_uninstall_plugin

    def search_enable_plugin(
        self,
        plugins: Iterable[str] | None = None,
        enable: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Enable or disable search plugin(s).

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).

        :param plugins: list of plugin names
        :param enable: Defaults to ``True`` if ``None`` or unset; use ``False`` to disable
        """
        data = {
            "names": self._list2string(plugins, "|"),
            "enable": True if enable is None else bool(enable),
        }
        self._post(
            _name=APINames.Search,
            _method="enablePlugin",
            data=data,
            version_introduced="2.1.1",
            **kwargs,
        )

    search_enablePlugin = search_enable_plugin

    def search_update_plugins(self, **kwargs: APIKwargsT) -> None:
        """
        Auto update search plugins.

        This method was introduced with qBittorrent v4.1.4 (Web API v2.1.1).
        """
        self._post(
            _name=APINames.Search,
            _method="updatePlugins",
            version_introduced="2.1.1",
            **kwargs,
        )

    search_updatePlugins = search_update_plugins


class SearchJobDictionary(ClientCache[SearchAPIMixIn], Dictionary[JsonValueT]):
    """Response for :meth:`~SearchAPIMixIn.search_start`"""

    def __init__(self, data: Mapping[str, JsonValueT], client: SearchAPIMixIn):
        self._search_job_id: int | None = cast(int, data.get("id", None))
        super().__init__(data=data, client=client)

    @wraps(SearchAPIMixIn.search_stop)
    def stop(self, **kwargs: APIKwargsT) -> None:
        self._client.search_stop(search_id=self._search_job_id, **kwargs)

    @wraps(SearchAPIMixIn.search_status)
    def status(self, **kwargs: APIKwargsT) -> SearchStatusesList:
        return self._client.search_status(search_id=self._search_job_id, **kwargs)

    @wraps(SearchAPIMixIn.search_results)
    def results(
        self,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        return self._client.search_results(
            limit=limit,
            offset=offset,
            search_id=self._search_job_id,
            **kwargs,
        )

    @wraps(SearchAPIMixIn.search_delete)
    def delete(self, **kwargs: APIKwargsT) -> None:
        return self._client.search_delete(search_id=self._search_job_id, **kwargs)


class Search(ClientCache[SearchAPIMixIn]):
    """
    Allows interaction with ``Search`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'search_' prepended)
        >>> # initiate searches and retrieve results
        >>> search_job = client.search.start(pattern="Ubuntu", plugins="all", category="all")
        >>> status = search_job.status()
        >>> results = search_job.result()
        >>> search_job.delete()
        >>> # inspect and manage plugins
        >>> plugins = client.search.plugins
        >>> cats = client.search.categories(plugin_name="...")
        >>> client.search.install_plugin(sources="...")
        >>> client.search.update_plugins()
    """

    @wraps(SearchAPIMixIn.search_start)
    def start(
        self,
        pattern: str | None = None,
        plugins: Iterable[str] | None = None,
        category: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchJobDictionary:
        return self._client.search_start(
            pattern=pattern,
            plugins=plugins,
            category=category,
            **kwargs,
        )

    @wraps(SearchAPIMixIn.search_stop)
    def stop(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.search_stop(search_id=search_id, **kwargs)

    @wraps(SearchAPIMixIn.search_status)
    def status(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchStatusesList:
        return self._client.search_status(search_id=search_id, **kwargs)

    @wraps(SearchAPIMixIn.search_results)
    def results(
        self,
        search_id: str | int | None = None,
        limit: str | int | None = None,
        offset: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchResultsDictionary:
        return self._client.search_results(
            search_id=search_id,
            limit=limit,
            offset=offset,
            **kwargs,
        )

    @wraps(SearchAPIMixIn.search_delete)
    def delete(
        self,
        search_id: str | int | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.search_delete(search_id=search_id, **kwargs)

    @wraps(SearchAPIMixIn.search_categories)
    def categories(
        self,
        plugin_name: str | None = None,
        **kwargs: APIKwargsT,
    ) -> SearchCategoriesList:
        return self._client.search_categories(plugin_name=plugin_name, **kwargs)

    @property
    @wraps(SearchAPIMixIn.search_plugins)
    def plugins(self) -> SearchPluginsList:
        return self._client.search_plugins()

    @wraps(SearchAPIMixIn.search_install_plugin)
    def install_plugin(
        self,
        sources: Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.search_install_plugin(sources=sources, **kwargs)

    installPlugin = install_plugin

    @wraps(SearchAPIMixIn.search_uninstall_plugin)
    def uninstall_plugin(
        self,
        sources: Iterable[str] | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.search_uninstall_plugin(sources=sources, **kwargs)

    uninstallPlugin = uninstall_plugin

    @wraps(SearchAPIMixIn.search_enable_plugin)
    def enable_plugin(
        self,
        plugins: Iterable[str] | None = None,
        enable: bool | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.search_enable_plugin(
            plugins=plugins,
            enable=enable,
            **kwargs,
        )

    enablePlugin = enable_plugin

    @wraps(SearchAPIMixIn.search_update_plugins)
    def update_plugins(self, **kwargs: APIKwargsT) -> None:
        return self._client.search_update_plugins(**kwargs)

    updatePlugins = update_plugins



from qbittorrentapi._version_support import Version
from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.app import ApplicationPreferencesDictionary
from qbittorrentapi.app import BuildInfoDictionary
from qbittorrentapi.app import NetworkInterface
from qbittorrentapi.app import NetworkInterfaceAddressList
from qbittorrentapi.app import NetworkInterfaceList
from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.client import Client
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import TorrentState
from qbittorrentapi.definitions import TorrentStates
from qbittorrentapi.definitions import TrackerStatus
from qbittorrentapi.exceptions import APIConnectionError
from qbittorrentapi.exceptions import APIError
from qbittorrentapi.exceptions import Conflict409Error
from qbittorrentapi.exceptions import FileError
from qbittorrentapi.exceptions import Forbidden403Error
from qbittorrentapi.exceptions import HTTP4XXError
from qbittorrentapi.exceptions import HTTP5XXError
from qbittorrentapi.exceptions import HTTP400Error
from qbittorrentapi.exceptions import HTTP401Error
from qbittorrentapi.exceptions import HTTP403Error
from qbittorrentapi.exceptions import HTTP404Error
from qbittorrentapi.exceptions import HTTP409Error
from qbittorrentapi.exceptions import HTTP415Error
from qbittorrentapi.exceptions import HTTP500Error
from qbittorrentapi.exceptions import HTTPError
from qbittorrentapi.exceptions import InternalServerError500Error
from qbittorrentapi.exceptions import InvalidRequest400Error
from qbittorrentapi.exceptions import LoginFailed
from qbittorrentapi.exceptions import MissingRequiredParameters400Error
from qbittorrentapi.exceptions import NotFound404Error
from qbittorrentapi.exceptions import TorrentFileError
from qbittorrentapi.exceptions import TorrentFileNotFoundError
from qbittorrentapi.exceptions import TorrentFilePermissionError
from qbittorrentapi.exceptions import Unauthorized401Error
from qbittorrentapi.exceptions import UnsupportedMediaType415Error
from qbittorrentapi.exceptions import UnsupportedQbittorrentVersion
from qbittorrentapi.log import LogAPIMixIn
from qbittorrentapi.log import LogEntry
from qbittorrentapi.log import LogMainList
from qbittorrentapi.log import LogPeer
from qbittorrentapi.log import LogPeersList
from qbittorrentapi.request import Request
from qbittorrentapi.rss import RSSAPIMixIn
from qbittorrentapi.rss import RSSitemsDictionary
from qbittorrentapi.rss import RSSRulesDictionary
from qbittorrentapi.search import SearchAPIMixIn
from qbittorrentapi.search import SearchCategoriesList
from qbittorrentapi.search import SearchCategory
from qbittorrentapi.search import SearchJobDictionary
from qbittorrentapi.search import SearchPlugin
from qbittorrentapi.search import SearchPluginsList
from qbittorrentapi.search import SearchResultsDictionary
from qbittorrentapi.search import SearchStatus
from qbittorrentapi.search import SearchStatusesList
from qbittorrentapi.sync import SyncAPIMixIn
from qbittorrentapi.sync import SyncMainDataDictionary
from qbittorrentapi.sync import SyncTorrentPeersDictionary
from qbittorrentapi.torrents import Tag
from qbittorrentapi.torrents import TagList
from qbittorrentapi.torrents import TorrentCategoriesDictionary
from qbittorrentapi.torrents import TorrentDictionary
from qbittorrentapi.torrents import TorrentFile
from qbittorrentapi.torrents import TorrentFilesList
from qbittorrentapi.torrents import TorrentInfoList
from qbittorrentapi.torrents import TorrentLimitsDictionary
from qbittorrentapi.torrents import TorrentPieceData
from qbittorrentapi.torrents import TorrentPieceInfoList
from qbittorrentapi.torrents import TorrentPropertiesDictionary
from qbittorrentapi.torrents import TorrentsAddPeersDictionary
from qbittorrentapi.torrents import TorrentsAPIMixIn
from qbittorrentapi.torrents import Tracker
from qbittorrentapi.torrents import TrackersList
from qbittorrentapi.torrents import WebSeed
from qbittorrentapi.torrents import WebSeedsList
from qbittorrentapi.transfer import TransferAPIMixIn
from qbittorrentapi.transfer import TransferInfoDictionary

__all__ = (
    "APIConnectionError",
    "APIError",
    "APINames",
    "AppAPIMixIn",
    "ApplicationPreferencesDictionary",
    "AuthAPIMixIn",
    "BuildInfoDictionary",
    "Client",
    "Conflict409Error",
    "FileError",
    "Forbidden403Error",
    "HTTP400Error",
    "HTTP401Error",
    "HTTP403Error",
    "HTTP404Error",
    "HTTP409Error",
    "HTTP415Error",
    "HTTP4XXError",
    "HTTP500Error",
    "HTTP5XXError",
    "HTTPError",
    "InternalServerError500Error",
    "InvalidRequest400Error",
    "LogAPIMixIn",
    "LogEntry",
    "LoginFailed",
    "LogMainList",
    "LogPeer",
    "LogPeersList",
    "MissingRequiredParameters400Error",
    "NetworkInterface",
    "NetworkInterfaceList",
    "NetworkInterfaceAddressList",
    "NotFound404Error",
    "Request",
    "RSSAPIMixIn",
    "RSSitemsDictionary",
    "RSSRulesDictionary",
    "SearchAPIMixIn",
    "SearchCategoriesList",
    "SearchCategory",
    "SearchJobDictionary",
    "SearchPlugin",
    "SearchPluginsList",
    "SearchResultsDictionary",
    "SearchStatus",
    "SearchStatus",
    "SearchStatusesList",
    "SyncAPIMixIn",
    "SyncMainDataDictionary",
    "SyncTorrentPeersDictionary",
    "Tag",
    "TagList",
    "TorrentCategoriesDictionary",
    "TorrentDictionary",
    "TorrentFile",
    "TorrentFileError",
    "TorrentFileNotFoundError",
    "TorrentFilePermissionError",
    "TorrentFilesList",
    "TorrentInfoList",
    "TorrentLimitsDictionary",
    "TorrentPieceData",
    "TorrentPieceInfoList",
    "TorrentPropertiesDictionary",
    "TorrentsAddPeersDictionary",
    "TorrentsAPIMixIn",
    "TorrentState",
    "TorrentStates",
    "Tracker",
    "TrackersList",
    "TrackerStatus",
    "TransferAPIMixIn",
    "TransferInfoDictionary",
    "Unauthorized401Error",
    "UnsupportedMediaType415Error",
    "UnsupportedQbittorrentVersion",
    "Version",
    "WebSeed",
    "WebSeedsList",
)



import errno
import platform
import sys
from time import sleep

import pytest
import requests

from qbittorrentapi import APINames
from qbittorrentapi._version_support import v
from qbittorrentapi.exceptions import Conflict409Error
from qbittorrentapi.exceptions import Forbidden403Error
from qbittorrentapi.exceptions import InvalidRequest400Error
from qbittorrentapi.exceptions import TorrentFileError
from qbittorrentapi.exceptions import TorrentFileNotFoundError
from qbittorrentapi.exceptions import TorrentFilePermissionError
from qbittorrentapi.torrents import TagList
from qbittorrentapi.torrents import TorrentCategoriesDictionary
from qbittorrentapi.torrents import TorrentFilesList
from qbittorrentapi.torrents import TorrentInfoList
from qbittorrentapi.torrents import TorrentLimitsDictionary
from qbittorrentapi.torrents import TorrentPieceInfoList
from qbittorrentapi.torrents import TorrentPropertiesDictionary
from qbittorrentapi.torrents import TorrentsAddPeersDictionary
from qbittorrentapi.torrents import TrackersList
from qbittorrentapi.torrents import WebSeedsList
from tests.conftest import ROOT_FOLDER_TORRENT_FILE
from tests.conftest import ROOT_FOLDER_TORRENT_HASH
from tests.conftest import TORRENT1_FILENAME
from tests.conftest import TORRENT1_HASH
from tests.conftest import TORRENT1_URL
from tests.conftest import TORRENT2_FILENAME
from tests.conftest import TORRENT2_HASH
from tests.conftest import TORRENT2_URL
from tests.conftest import new_torrent_standalone
from tests.utils import check
from tests.utils import mkpath
from tests.utils import retry


def disable_queueing(client):
    if client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=False))


def enable_queueing(client):
    if not client.app.preferences.queueing_enabled:
        client.app.set_preferences(dict(queueing_enabled=True))


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    all_dotted_methods = {
        meth
        for namespace in [APINames.Torrents, "torrent_tags", "torrent_categories"]
        for meth in dir(getattr(client, namespace))
    }

    for meth in [meth for meth in dir(client) if meth.startswith("torrents_")]:
        assert meth.removeprefix("torrents_") in all_dotted_methods


# something was wrong with torrents_add on v2.0.0 (the initial version)
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "add_func, delete_func",
    [("torrents_add", "torrents_delete"), ("torrents.add", "torrents.delete")],
)
def test_add_delete(client, add_func, delete_func, tmp_path):
    def download_file(url, filename=None, return_bytes=False):
        max_attempts = 3
        for attempt in range(max_attempts):
            try:
                with requests.get(url, timeout=30) as r:
                    r.raise_for_status()
                    if return_bytes:
                        return r.content
                    with open(mkpath(tmp_path, filename), "wb") as f:
                        for chunk in r.iter_content(chunk_size=1024):
                            f.write(chunk)
            except Exception if attempt < (max_attempts - 1) else ZeroDivisionError:
                pass  # throw away errors until we hit the retry limit
            else:
                return
        raise Exception("Download failed: %s" % url)

    def delete():
        client.func(delete_func)(delete_files=True, torrent_hashes=TORRENT1_HASH)
        client.func(delete_func)(delete_files=True, torrent_hashes=TORRENT2_HASH)
        check(
            lambda: [t.hash for t in client.torrents_info()],
            TORRENT2_HASH,
            reverse=True,
            negate=True,
        )

    def check_torrents_added(f):
        def inner(**kwargs):
            try:
                f(**kwargs)
                check(
                    lambda: [t.hash for t in client.torrents_info()],
                    TORRENT1_HASH,
                    reverse=True,
                )
                if kwargs.get("single", False) is False:
                    check(
                        lambda: [t.hash for t in client.torrents_info()],
                        TORRENT2_HASH,
                        reverse=True,
                    )
            finally:
                delete()

        return inner

    @retry()
    @check_torrents_added
    def add_by_filename(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)
        files = (
            mkpath(tmp_path, TORRENT1_FILENAME),
            mkpath(tmp_path, TORRENT2_FILENAME),
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_filename_dict(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)

        if single:
            assert (
                client.func(add_func)(
                    torrent_files={
                        TORRENT1_FILENAME: mkpath(tmp_path, TORRENT1_FILENAME)
                    }
                )
                == "Ok."
            )
        else:
            files = {
                TORRENT1_FILENAME: mkpath(tmp_path, TORRENT1_FILENAME),
                TORRENT2_FILENAME: mkpath(tmp_path, TORRENT2_FILENAME),
            }
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_filehandles(single):
        download_file(url=TORRENT1_URL, filename=TORRENT1_FILENAME)
        download_file(url=TORRENT2_URL, filename=TORRENT2_FILENAME)
        files = (
            open(mkpath(tmp_path, TORRENT1_FILENAME), "rb"),
            open(mkpath(tmp_path, TORRENT2_FILENAME), "rb"),
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

        for file in files:
            file.close()

    @retry()
    @check_torrents_added
    def add_by_bytes(single):
        files = (
            download_file(TORRENT1_URL, return_bytes=True),
            download_file(TORRENT2_URL, return_bytes=True),
        )

        if single:
            assert client.func(add_func)(torrent_files=files[0]) == "Ok."
        else:
            assert client.func(add_func)(torrent_files=files) == "Ok."

    @retry()
    @check_torrents_added
    def add_by_url(single):
        urls = (TORRENT1_URL, TORRENT2_URL)

        if single:
            client.func(add_func)(urls=urls[0])
        else:
            client.func(add_func)(urls=urls)

    add_by_filename(single=False)
    add_by_filename(single=True)
    add_by_filename_dict(single=False)
    add_by_filename_dict(single=True)
    add_by_url(single=False)
    add_by_url(single=True)
    add_by_filehandles(single=False)
    add_by_filehandles(single=True)
    add_by_bytes(single=False)
    add_by_bytes(single=True)


def test_add_torrent_file_fail(client, monkeypatch):
    with pytest.raises(TorrentFileNotFoundError):
        client.torrents_add(torrent_files="/tmp/asdfasdfasdfasdf")

    with pytest.raises(TorrentFilePermissionError):
        client.torrents_add(torrent_files="/etc/shadow")

    if platform.python_implementation() == "CPython":
        with pytest.raises(TorrentFileError):

            def fake_open(*arg, **kwargs):
                raise OSError(errno.ENODEV)

            with monkeypatch.context() as m:
                m.setitem(__builtins__, "open", fake_open)
                client.torrents_add(torrent_files="/etc/hosts")


@pytest.mark.parametrize("keep_root_folder", [True, False, None])
@pytest.mark.parametrize(
    "content_layout", [None, "Original", "Subfolder", "NoSubfolder"]
)
def test_add_options(client, api_version, keep_root_folder, content_layout, tmp_path):
    if v(api_version) >= v("2.3.0"):
        client.torrents_create_tags("option-tag")
    new_torrent = new_torrent_standalone(
        client=client,
        torrent_files=ROOT_FOLDER_TORRENT_FILE,
        torrent_hash=ROOT_FOLDER_TORRENT_HASH,
        save_path=mkpath(tmp_path, "test_download"),
        category="test_category",
        is_paused=True,
        upload_limit=1024,
        download_limit=2048,
        is_sequential_download=True,
        is_first_last_piece_priority=True,
        is_root_folder=keep_root_folder,
        rename="this is a new name for the torrent",
        use_auto_torrent_management=False,
        tags="option-tag",
        content_layout=content_layout,
        ratio_limit=2,
        seeding_time_limit=120,
    )

    with new_torrent as torrent:
        check(lambda: torrent.info.category, "test_category")
        check(
            lambda: torrent.info.state,
            ("pausedDL", "checkingResumeData"),
            reverse=True,
            any=True,
        )
        check(lambda: mkpath(torrent.info.save_path), mkpath(tmp_path, "test_download"))
        check(lambda: torrent.info.up_limit, 1024)
        check(lambda: torrent.info.dl_limit, 2048)
        check(lambda: torrent.info.seq_dl, True)
        if v(api_version) >= v("2.0.1"):
            check(lambda: torrent.info.f_l_piece_prio, True)
        if content_layout is None:
            check(
                lambda: torrent.files[0]["name"].startswith("root_folder"),
                keep_root_folder in {True, None},
            )
        check(lambda: torrent.info.name, "this is a new name for the torrent")
        check(lambda: torrent.info.auto_tmm, False)
        if v(api_version) >= v("2.6.2"):
            check(lambda: torrent.info.tags, "option-tag")

        if v(api_version) >= v("2.7"):
            # after web api v2.7...root dir is driven by content_layout
            if content_layout is None:
                should_root_dir_exists = keep_root_folder in {None, True}
            else:
                should_root_dir_exists = content_layout in {"Original", "Subfolder"}
        else:
            # before web api v2.7...it is driven by is_root_folder
            if content_layout is not None and keep_root_folder is None:
                should_root_dir_exists = content_layout in {"Original", "Subfolder"}
            else:
                should_root_dir_exists = keep_root_folder in {None, True}
        check(
            lambda: any(f["name"].startswith("root_folder") for f in torrent.files),
            should_root_dir_exists,
        )

        if v(api_version) >= v("2.8.1"):
            check(lambda: torrent.info.ratio_limit, 2)
            check(lambda: torrent.info.seeding_time_limit, 120)


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("use_download_path", [None, True, False])
def test_torrents_add_download_path(client, use_download_path, tmp_path):
    client.torrents_delete(torrent_hashes=ROOT_FOLDER_TORRENT_HASH, delete_files=True)
    save_path = mkpath(tmp_path, "down_path_save_path_test")
    download_path = mkpath(tmp_path, "down_path_test")
    new_torrent = new_torrent_standalone(
        client=client,
        torrent_hash=ROOT_FOLDER_TORRENT_HASH,
        torrent_files=ROOT_FOLDER_TORRENT_FILE,
        download_path=download_path,
        use_download_path=use_download_path,
        test_download_limit=1024,
        save_path=save_path,
    )

    with new_torrent as torrent:
        if use_download_path is False:
            check(
                lambda: mkpath(torrent.info.download_path), download_path, negate=True
            )
        else:
            check(lambda: mkpath(torrent.info.download_path), download_path)


@pytest.mark.skipif_before_api_version("2.9.3")
@pytest.mark.parametrize("count_func", ["torrents_count", "torrents.count"])
def test_count(client, count_func):
    assert client.func(count_func)() == 1


@pytest.mark.skipif_after_api_version("2.9.3")
@pytest.mark.parametrize("count_func", ["torrents_count", "torrents.count"])
def test_count_not_implemented(client, count_func):
    with pytest.raises(NotImplementedError):
        assert client.func(count_func)() == 1


@pytest.mark.parametrize(
    "properties_func", ["torrents_properties", "torrents.properties"]
)
def test_properties(client, orig_torrent, properties_func):
    props = client.func(properties_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(props, TorrentPropertiesDictionary)


@pytest.mark.parametrize("trackers_func", ["torrents_trackers", "torrents.trackers"])
def test_trackers(client, orig_torrent, trackers_func):
    trackers = client.func(trackers_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(trackers, TrackersList)


@pytest.mark.parametrize("trackers_func", ["torrents_trackers", "torrents.trackers"])
def test_trackers_slice(client, orig_torrent, trackers_func):
    trackers = client.func(trackers_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(trackers[1:2], TrackersList)


@pytest.mark.parametrize("webseeds_func", ["torrents_webseeds", "torrents.webseeds"])
def test_webseeds(client, orig_torrent, webseeds_func):
    web_seeds = client.func(webseeds_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(web_seeds, WebSeedsList)


@pytest.mark.parametrize("webseeds_func", ["torrents_webseeds", "torrents.webseeds"])
def test_webseeds_slice(client, orig_torrent, webseeds_func):
    web_seeds = client.func(webseeds_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(web_seeds[1:2], WebSeedsList)


@pytest.mark.parametrize("files_func", ["torrents_files", "torrents.files"])
def test_files(client, orig_torrent, files_func):
    files = client.func(files_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(files, TorrentFilesList)
    assert "availability" in files[0]
    assert all(file["id"] == file["index"] for file in files)


@pytest.mark.parametrize("files_func", ["torrents_files", "torrents.files"])
def test_files_slice(client, orig_torrent, files_func):
    files = client.func(files_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(files[1:2], TorrentFilesList)


@pytest.mark.parametrize(
    "piece_state_func",
    [
        "torrents_piece_states",
        "torrents_pieceStates",
        "torrents.piece_states",
        "torrents.pieceStates",
    ],
)
def test_piece_states(client, orig_torrent, piece_state_func):
    piece_states = client.func(piece_state_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_states, TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_state_func",
    [
        "torrents_piece_states",
        "torrents_pieceStates",
        "torrents.piece_states",
        "torrents.pieceStates",
    ],
)
def test_piece_states_slice(client, orig_torrent, piece_state_func):
    piece_states = client.func(piece_state_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_states[1:2], TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_hashes_func",
    [
        "torrents_piece_hashes",
        "torrents_pieceHashes",
        "torrents.piece_hashes",
        "torrents.pieceHashes",
    ],
)
def test_piece_hashes(client, orig_torrent, piece_hashes_func):
    piece_hashes = client.func(piece_hashes_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_hashes, TorrentPieceInfoList)


@pytest.mark.parametrize(
    "piece_hashes_func",
    [
        "torrents_piece_hashes",
        "torrents_pieceHashes",
        "torrents.piece_hashes",
        "torrents.pieceHashes",
    ],
)
def test_piece_hashes_slice(client, orig_torrent, piece_hashes_func):
    piece_hashes = client.func(piece_hashes_func)(torrent_hash=orig_torrent.hash)
    assert isinstance(piece_hashes[1:2], TorrentPieceInfoList)


@pytest.mark.parametrize("trackers", ["127.0.0.1", ["127.0.0.2", "127.0.0.3"]])
@pytest.mark.parametrize(
    "add_trackers_func",
    [
        "torrents_add_trackers",
        "torrents_addTrackers",
        "torrents.add_trackers",
        "torrents.addTrackers",
    ],
)
def test_add_trackers(client, trackers, new_torrent, add_trackers_func):
    client.func(add_trackers_func)(torrent_hash=new_torrent.hash, urls=trackers)
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize(
    "edit_trackers_func",
    [
        "torrents_edit_tracker",
        "torrents_editTracker",
        "torrents.edit_tracker",
        "torrents.editTracker",
    ],
)
def test_edit_tracker(client, orig_torrent, edit_trackers_func):
    orig_torrent.add_trackers("127.1.0.1")
    client.func(edit_trackers_func)(
        torrent_hash=orig_torrent.hash,
        original_url="127.1.0.1",
        new_url="127.1.0.2",
    )
    check(lambda: (t.url for t in orig_torrent.trackers), "127.1.0.2", reverse=True)
    client.torrents_remove_trackers(torrent_hash=orig_torrent.hash, urls="127.1.0.2")


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize(
    "edit_trackers_func",
    [
        "torrents_edit_tracker",
        "torrents_editTracker",
        "torrents.edit_tracker",
        "torrents.editTracker",
    ],
)
def test_edit_tracker_not_implemented(client, orig_torrent, edit_trackers_func):
    with pytest.raises(NotImplementedError):
        client.func(edit_trackers_func)()


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize(
    "trackers",
    [
        ["127.2.0.1"],
        ["127.2.0.2", "127.2.0.3"],
    ],
)
@pytest.mark.parametrize(
    "remove_trackers_func",
    [
        "torrents_remove_trackers",
        "torrents_removeTrackers",
        "torrents.remove_trackers",
        "torrents.removeTrackers",
    ],
)
def test_remove_trackers(client, trackers, orig_torrent, remove_trackers_func):
    orig_torrent.add_trackers(trackers)
    client.func(remove_trackers_func)(torrent_hash=orig_torrent.hash, urls=trackers)
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize(
    "remove_trackers_func",
    [
        "torrents_remove_trackers",
        "torrents_removeTrackers",
        "torrents.remove_trackers",
        "torrents.removeTrackers",
    ],
)
def test_remove_trackers_not_implemented(client, orig_torrent, remove_trackers_func):
    with pytest.raises(NotImplementedError):
        client.func(remove_trackers_func)()


@pytest.mark.parametrize(
    "file_prio_func",
    [
        "torrents_file_priority",
        "torrents_filePrio",
        "torrents.file_priority",
        "torrents.filePrio",
    ],
)
def test_file_priority(client, orig_torrent, file_prio_func):
    client.func(file_prio_func)(torrent_hash=orig_torrent.hash, file_ids=0, priority=6)
    check(lambda: orig_torrent.files[0].priority, 6)
    client.func(file_prio_func)(torrent_hash=orig_torrent.hash, file_ids=0, priority=7)
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("new_name", ["new name 2", "new_name_2"])
@pytest.mark.parametrize("rename_func", ["torrents_rename", "torrents.rename"])
def test_rename(client, new_torrent, new_name, rename_func):
    client.func(rename_func)(torrent_hash=new_torrent.hash, new_torrent_name=new_name)
    check(lambda: new_torrent.info.name.replace("+", " "), new_name)


@pytest.mark.skipif_before_api_version("2.4.0")
@pytest.mark.parametrize("new_name", ["new name file 2", "new_name_file_2"])
@pytest.mark.parametrize(
    "rename_file_func",
    [
        "torrents_rename_file",
        "torrents_renameFile",
        "torrents.rename_file",
        "torrents.renameFile",
    ],
)
def test_rename_file(
    client,
    new_torrent,
    new_name,
    rename_file_func,
):
    # pre-v4.3.3 rename_file signature
    client.func(rename_file_func)(
        torrent_hash=new_torrent.hash, file_id=0, new_file_name=new_name
    )
    check(lambda: new_torrent.files[0].name.replace("+", " "), new_name)
    # test invalid file ID is rejected
    with pytest.raises(Conflict409Error):
        client.func(rename_file_func)(
            torrent_hash=new_torrent.hash, file_id=10, new_file_name=new_name
        )
    # post-v4.3.3 rename_file signature
    new_new_name = new_name + "NEW"
    client.func(rename_file_func)(
        torrent_hash=new_torrent.hash,
        old_path=new_torrent.files[0].name,
        new_path=new_new_name,
    )
    check(lambda: new_torrent.files[0].name.replace("+", " "), new_new_name)
    # test invalid old_path is rejected
    with pytest.raises(Conflict409Error):
        client.func(rename_file_func)(
            torrent_hash=new_torrent.hash, old_path="asdf", new_path="xcvb"
        )


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize(
    "rename_file_func",
    [
        "torrents_rename_file",
        "torrents_renameFile",
        "torrents.rename_file",
        "torrents.renameFile",
    ],
)
def test_rename_file_not_implemented(
    client,
    new_torrent,
    rename_file_func,
):
    with pytest.raises(NotImplementedError):
        client.func(rename_file_func)()


@pytest.mark.skipif_before_api_version("2.7")
@pytest.mark.parametrize("new_name", ["asdf zxcv", "asdf_zxcv"])
@pytest.mark.parametrize(
    "rename_folder_func",
    [
        "torrents_rename_folder",
        "torrents_renameFolder",
        "torrents.rename_folder",
        "torrents.renameFolder",
    ],
)
def test_rename_folder(client, app_version, new_torrent, new_name, rename_folder_func):
    if v(app_version) >= v("v4.3.3"):
        # move the file in to a new folder
        orig_file_path = new_torrent.files[0].name
        new_folder = "qwer"
        client.torrents_rename_file(
            torrent_hash=new_torrent.hash,
            old_path=orig_file_path,
            new_path=new_folder + "/" + orig_file_path,
        )

        # wait for the folder to be renamed
        check(
            lambda: [f.name.split("/")[0] for f in new_torrent.files],
            new_folder,
            reverse=True,
        )

        # test rename that new folder
        client.func(rename_folder_func)(
            torrent_hash=new_torrent.hash,
            old_path=new_folder,
            new_path=new_name,
        )
        check(
            lambda: new_torrent.files[0].name.replace("+", " "),
            new_name + "/" + orig_file_path,
        )
    elif v(app_version) >= v("v4.3.2"):
        with pytest.raises(NotImplementedError):
            client.func(rename_folder_func)()


@pytest.mark.skipif_after_api_version("2.7")
@pytest.mark.parametrize(
    "rename_folder_func",
    [
        "torrents_rename_folder",
        "torrents_renameFolder",
        "torrents.rename_folder",
        "torrents.renameFolder",
    ],
)
def test_rename_folder_not_implemented(client, rename_folder_func):
    with pytest.raises(NotImplementedError):
        client.func(rename_folder_func)()


@pytest.mark.skipif_before_api_version("2.8.14")
@pytest.mark.parametrize("export_func", ["torrents_export", "torrents.export"])
def test_export(client, orig_torrent, export_func):
    assert isinstance(client.func(export_func)(torrent_hash=orig_torrent.hash), bytes)


@pytest.mark.skipif_after_api_version("2.8.14")
@pytest.mark.parametrize("export_func", ["torrents_export", "torrents.export"])
def test_export_not_implemented(client, export_func):
    with pytest.raises(NotImplementedError):
        client.func(export_func)()


@pytest.mark.parametrize("info_func", ["torrents_info", "torrents.info"])
def test_torrents_info(client, info_func):
    assert isinstance(client.func(info_func)(), TorrentInfoList)
    if "." in info_func:
        assert isinstance(client.func(info_func).all(), TorrentInfoList)
        assert isinstance(client.func(info_func).downloading(), TorrentInfoList)
        assert isinstance(client.func(info_func).seeding(), TorrentInfoList)
        assert isinstance(client.func(info_func).completed(), TorrentInfoList)
        assert isinstance(client.func(info_func).paused(), TorrentInfoList)
        assert isinstance(client.func(info_func).active(), TorrentInfoList)
        assert isinstance(client.func(info_func).inactive(), TorrentInfoList)
        assert isinstance(client.func(info_func).resumed(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled_uploading(), TorrentInfoList)
        assert isinstance(client.func(info_func).stalled_downloading(), TorrentInfoList)
        assert isinstance(client.func(info_func).checking(), TorrentInfoList)
        assert isinstance(client.func(info_func).moving(), TorrentInfoList)
        assert isinstance(client.func(info_func).errored(), TorrentInfoList)


def test_torrents_info_slice(client):
    assert isinstance(client.torrents_info()[1:2], TorrentInfoList)


@pytest.mark.skipif_before_api_version("2.8.3")
@pytest.mark.parametrize("info_func", ["torrents_info", "torrents.info"])
def test_torrents_info_tag(client, new_torrent, info_func):
    tag_name = "tag_filter_name"
    client.torrents_add_tags(tags=tag_name, torrent_hashes=new_torrent.hash)
    torrents = client.func(info_func)(torrent_hashes=new_torrent.hash, tag=tag_name)
    assert new_torrent.hash in {t.hash for t in torrents}


# test fails on 4.1.0 release
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "pause_func, resume_func",
    [("torrents_pause", "torrents_resume"), ("torrents.pause", "torrents.resume")],
)
def test_pause_resume(client, new_torrent, pause_func, resume_func):
    client.func(pause_func)(torrent_hashes=new_torrent.hash)
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        True,
    )

    client.func(resume_func)(torrent_hashes=new_torrent.hash)
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        False,
    )


def test_action_for_all_torrents(client):
    client.torrents.resume.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrent_hashes=torrent.hash)[0].state,
            {"pausedDL"},
            negate=True,
        )
    client.torrents.pause.all()
    for torrent in client.torrents.info():
        check(
            lambda: client.torrents_info(torrent_hashes=torrent.hash)[0].state,
            {"stalledDL", "pausedDL"},
            any=True,
        )


@pytest.mark.parametrize("recheck_func", ["torrents_recheck", "torrents.recheck"])
def test_recheck(client, orig_torrent, recheck_func):
    client.func(recheck_func)(torrent_hashes=orig_torrent.hash)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "reannounce_func", ["torrents_reannounce", "torrents.reannounce"]
)
def test_reannounce(client, orig_torrent, reannounce_func):
    client.func(reannounce_func)(torrent_hashes=orig_torrent.hash)


@pytest.mark.skipif_after_api_version("2.0.2")
@pytest.mark.parametrize(
    "reannounce_func", ["torrents_reannounce", "torrents.reannounce"]
)
def test_reannounce_not_implemented(client, reannounce_func):
    with pytest.raises(NotImplementedError):
        client.func(reannounce_func)()


# priority doesn't seem to work on v4.1.0
@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func",
    [
        (
            "torrents_increase_priority",
            "torrents_decrease_priority",
            "torrents_top_priority",
            "torrents_bottom_priority",
        ),
        (
            "torrents_increasePrio",
            "torrents_decreasePrio",
            "torrents_topPrio",
            "torrents_bottomPrio",
        ),
    ],
)
def test_priority(
    client, new_torrent, inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func
):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        client.func(inc_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(dec_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(top_prio_func)(torrent_hashes=new_torrent.hash)
    with pytest.raises(Conflict409Error):
        client.func(bottom_prio_func)(torrent_hashes=new_torrent.hash)

    enable_queueing(client)

    @retry()
    def test1(current_priority):
        client.func(inc_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry()
    def test2(current_priority):
        client.func(dec_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    @retry()
    def test3(current_priority):
        client.func(top_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority < current_priority, True)

    @retry()
    def test4(current_priority):
        client.func(bottom_prio_func)(torrent_hashes=new_torrent.hash)
        check(lambda: new_torrent.info.priority > current_priority, True)

    test1(current_priority=new_torrent.info.priority)
    test2(current_priority=new_torrent.info.priority)
    test3(current_priority=new_torrent.info.priority)
    test4(current_priority=new_torrent.info.priority)


@pytest.mark.parametrize(
    "set_down_limit_func, down_limit_func",
    [
        ("torrents_set_download_limit", "torrents_download_limit"),
        ("torrents_setDownloadLimit", "torrents_downloadLimit"),
        ("torrents.set_download_limit", "torrents.download_limit"),
        ("torrents.setDownloadLimit", "torrents.downloadLimit"),
    ],
)
def test_download_limit(client, orig_torrent, set_down_limit_func, down_limit_func):
    orig_download_limit = client.func(down_limit_func)(
        torrent_hashes=orig_torrent.hash
    )[orig_torrent.hash]

    client.func(set_down_limit_func)(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        client.func(down_limit_func)(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: client.func(down_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset download limit
    client.func(set_down_limit_func)(
        torrent_hashes=orig_torrent.hash, limit=orig_download_limit
    )
    check(
        lambda: client.func(down_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_download_limit,
    )


@pytest.mark.parametrize(
    "set_up_limit_func, up_limit_func",
    [
        ("torrents_set_upload_limit", "torrents_upload_limit"),
        ("torrents_setUploadLimit", "torrents_uploadLimit"),
        ("torrents.set_upload_limit", "torrents.upload_limit"),
        ("torrents.setUploadLimit", "torrents.uploadLimit"),
    ],
)
def test_upload_limit(client, orig_torrent, set_up_limit_func, up_limit_func):
    orig_upload_limit = client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
        orig_torrent.hash
    ]

    client.func(set_up_limit_func)(torrent_hashes=orig_torrent.hash, limit=100)
    assert isinstance(
        client.func(up_limit_func)(torrent_hashes=orig_torrent.hash),
        TorrentLimitsDictionary,
    )
    check(
        lambda: client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        100,
    )

    # reset upload limit
    client.func(set_up_limit_func)(
        torrent_hashes=orig_torrent.hash, limit=orig_upload_limit
    )
    check(
        lambda: client.func(up_limit_func)(torrent_hashes=orig_torrent.hash)[
            orig_torrent.hash
        ],
        orig_upload_limit,
    )


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func",
    [
        "torrents_set_share_limits",
        "torrents_setShareLimits",
        "torrents.set_share_limits",
        "torrents.setShareLimits",
    ],
)
def test_set_share_limits(client, orig_torrent, set_share_limits_func):
    client.func(set_share_limits_func)(
        ratio_limit=2,
        seeding_time_limit=5,
        inactive_seeding_time_limit=8,
        torrent_hashes=orig_torrent.hash,
    )
    check(lambda: orig_torrent.info.max_ratio, 2)
    check(lambda: orig_torrent.info.max_seeding_time, 5)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 8)

    client.func(set_share_limits_func)(
        ratio_limit=3,
        seeding_time_limit=6,
        inactive_seeding_time_limit=9,
        torrent_hashes=orig_torrent.hash,
    )
    check(lambda: orig_torrent.info.max_ratio, 3)
    check(lambda: orig_torrent.info.max_seeding_time, 6)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 9)


@pytest.mark.skipif_after_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func",
    [
        "torrents_set_share_limits",
        "torrents_setShareLimits",
        "torrents.set_share_limits",
        "torrents.setShareLimits",
    ],
)
def test_set_share_limits_not_implemented(client, set_share_limits_func):
    with pytest.raises(NotImplementedError):
        client.func(set_share_limits_func)()


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "set_loc_func",
    [
        "torrents_set_location",
        "torrents_setLocation",
        "torrents.set_location",
        "torrents.setLocation",
    ],
)
def test_set_location(client, app_version, new_torrent, set_loc_func, tmp_path):
    # stopped erroring when the write check was removed for API
    if v(app_version) < v("v4.5.2"):
        with pytest.raises(Forbidden403Error):
            client.func(set_loc_func)(location="/etc/", torrent_hashes=new_torrent.hash)

    sleep(0.5)
    loc = mkpath(tmp_path, "1")
    client.func(set_loc_func)(location=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_save_path_func",
    [
        "torrents_set_save_path",
        "torrents_setSavePath",
        "torrents.set_save_path",
        "torrents.setSavePath",
    ],
)
def test_set_save_path(client, new_torrent, set_save_path_func, tmp_path):
    with pytest.raises(Forbidden403Error):
        client.func(set_save_path_func)(
            save_path="/etc/", torrent_hashes=new_torrent.hash
        )
    with pytest.raises(Conflict409Error):
        client.func(set_save_path_func)(
            save_path="/etc/asdf", torrent_hashes=new_torrent.hash
        )

    loc = mkpath(tmp_path, "savepath1")
    client.func(set_save_path_func)(save_path=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), loc, any=True)


@pytest.mark.skipif_after_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_save_path_func",
    [
        "torrents_set_save_path",
        "torrents_setSavePath",
        "torrents.set_save_path",
        "torrents.setSavePath",
    ],
)
def test_set_save_path_not_implemented(client, set_save_path_func):
    with pytest.raises(NotImplementedError):
        client.func(set_save_path_func)()


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_down_path_func",
    [
        "torrents_set_download_path",
        "torrents_setDownloadPath",
        "torrents.set_download_path",
        "torrents.setDownloadPath",
    ],
)
def test_set_download_path(client, new_torrent, set_down_path_func, tmp_path):
    with pytest.raises(Forbidden403Error):
        client.func(set_down_path_func)(
            download_path="/etc/", torrent_hashes=new_torrent.hash
        )
    with pytest.raises(Conflict409Error):
        client.func(set_down_path_func)(
            download_path="/etc/asdf", torrent_hashes=new_torrent.hash
        )

    loc = mkpath(tmp_path, "savepath1")
    client.func(set_down_path_func)(download_path=loc, torrent_hashes=new_torrent.hash)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.download_path), loc, any=True)


@pytest.mark.skipif_after_api_version("2.8.4")
@pytest.mark.parametrize(
    "set_down_path_func",
    [
        "torrents_set_download_path",
        "torrents_setDownloadPath",
        "torrents.set_download_path",
        "torrents.setDownloadPath",
    ],
)
def test_set_download_path_not_implemented(client, new_torrent, set_down_path_func):
    with pytest.raises(NotImplementedError):
        client.func(set_down_path_func)()


@pytest.mark.parametrize(
    "set_cat_func",
    [
        "torrents_set_category",
        "torrents_setCategory",
        "torrents.set_category",
        "torrents.setCategory",
    ],
)
@pytest.mark.parametrize("name", ["awesome cat", "awesome_cat"])
def test_set_category(client, orig_torrent, set_cat_func, name):
    with pytest.raises(Conflict409Error):
        client.func(set_cat_func)(
            category="/!@#$%^&*(", torrent_hashes=orig_torrent.hash
        )

    client.torrents_create_category(name=name)
    try:
        client.func(set_cat_func)(category=name, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.parametrize(
    "set_auto_mgmt_func",
    [
        "torrents_set_auto_management",
        "torrents_setAutoManagement",
        "torrents.set_auto_management",
        "torrents.setAutoManagement",
    ],
)
def test_torrents_set_auto_management(client, orig_torrent, set_auto_mgmt_func):
    current_setting = orig_torrent.info.auto_tmm
    client.func(set_auto_mgmt_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.auto_tmm, (not current_setting))
    client.func(set_auto_mgmt_func)(
        enable=False, torrent_hashes=orig_torrent.hash
    )  # leave on False


@pytest.mark.parametrize(
    "toggle_seq_down_func",
    [
        "torrents_toggle_sequential_download",
        "torrents_toggleSequentialDownload",
        "torrents.toggle_sequential_download",
        "torrents.toggleSequentialDownload",
    ],
)
def test_toggle_sequential_download(client, orig_torrent, toggle_seq_down_func):
    current_setting = orig_torrent.info.seq_dl
    client.func(toggle_seq_down_func)(torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.seq_dl, not current_setting)


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "toggle_piece_prio_func",
    [
        "torrents_toggle_first_last_piece_priority",
        "torrents_toggleFirstLastPiecePrio",
        "torrents.toggle_first_last_piece_priority",
        "torrents.toggleFirstLastPiecePrio",
    ],
)
def test_toggle_first_last_piece_priority(client, new_torrent, toggle_piece_prio_func):
    current_setting = new_torrent.info.f_l_piece_prio
    client.func(toggle_piece_prio_func)(torrent_hashes=new_torrent.hash)
    check(lambda: new_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize(
    "set_force_start_func",
    [
        "torrents_set_force_start",
        "torrents_setForceStart",
        "torrents.set_force_start",
        "torrents.setForceStart",
    ],
)
def test_set_force_start(client, orig_torrent, set_force_start_func):
    current_setting = orig_torrent.info.force_start
    client.func(set_force_start_func)(
        enable=(not current_setting), torrent_hashes=orig_torrent.hash
    )
    check(lambda: orig_torrent.info.force_start, not current_setting)


@pytest.mark.parametrize(
    "set_super_seeding_func",
    [
        "torrents_set_super_seeding",
        "torrents_setSuperSeeding",
        "torrents.set_super_seeding",
        "torrents.setSuperSeeding",
    ],
)
def test_set_super_seeding(client, orig_torrent, set_super_seeding_func):
    client.func(set_super_seeding_func)(enable=False, torrent_hashes=orig_torrent.hash)
    check(lambda: orig_torrent.info.force_start, False)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_peers_func",
    [
        "torrents_add_peers",
        "torrents_addPeers",
        "torrents.add_peers",
        "torrents.addPeers",
    ],
)
@pytest.mark.parametrize(
    "peers", ["127.0.0.1:5000", ["127.0.0.1:5000", "127.0.0.2:5000"], "127.0.0.1"]
)
def test_torrents_add_peers(client, orig_torrent, add_peers_func, peers):
    if all(":" not in p for p in peers):
        with pytest.raises(InvalidRequest400Error):
            client.func(add_peers_func)(peers=peers, torrent_hashes=orig_torrent.hash)
    else:
        p = client.func(add_peers_func)(peers=peers, torrent_hashes=orig_torrent.hash)
        assert isinstance(p, TorrentsAddPeersDictionary)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_peers_func",
    [
        "torrents_add_peers",
        "torrents_addPeers",
        "torrents.add_peers",
        "torrents.addPeers",
    ],
)
def test_torrents_add_peers_not_implemented(client, add_peers_func):
    with pytest.raises(NotImplementedError):
        client.func(add_peers_func)()


def _categories_save_path_key(api_version):
    """With qBittorrent 4.4.0 (Web API 2.8.4), the key in the category definition
    returned changed from savePath to save_path...."""
    if v(api_version) == v("2.8.4"):
        return "save_path"
    return "savePath"


@pytest.mark.skipif_before_api_version("2.1.1")
def test_categories1(client):
    assert isinstance(client.torrents_categories(), TorrentCategoriesDictionary)


@pytest.mark.skipif_after_api_version("2.1.1")
def test_categories1_not_implemented(client, api_version):
    with pytest.raises(NotImplementedError):
        client.torrents_categories()


@pytest.mark.skipif_before_api_version("2.1.1")
def test_categories2(client, api_version, tmp_path):
    save_path_key = _categories_save_path_key(api_version)
    name = "new_category"
    client.torrent_categories.categories = {"name": name, save_path_key: tmp_path}
    assert name in client.torrent_categories.categories
    client.torrent_categories.categories = {
        "name": name,
        save_path_key: mkpath(tmp_path, "new"),
    }
    assert mkpath(client.torrent_categories.categories[name][save_path_key]) == mkpath(
        tmp_path, "new"
    )
    client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_after_api_version("2.1.1")
def test_categories2_not_implemented(client):
    with pytest.raises(NotImplementedError):
        _ = client.torrent_categories.categories


@pytest.mark.parametrize(
    "create_cat_func",
    [
        "torrents_create_category",
        "torrents_createCategory",
        "torrent_categories.create_category",
        "torrent_categories.createCategory",
    ],
)
@pytest.mark.parametrize("filepath", [None, "", "/tmp/"])
@pytest.mark.parametrize("name", ["name", "name 1"])
@pytest.mark.parametrize("enable_download_path", [None, True, False])
def test_create_categories(
    client,
    api_version,
    orig_torrent,
    create_cat_func,
    filepath,
    name,
    enable_download_path,
):
    save_path = download_path = filepath
    if filepath:
        save_path += "save"
        download_path += "download"

    try:
        client.func(create_cat_func)(
            name=name,
            save_path=save_path,
            download_path=download_path,
            enable_download_path=enable_download_path,
        )
        client.torrents_set_category(torrent_hashes=orig_torrent.hash, category=name)
        check(lambda: orig_torrent.info.category.replace("+", " "), name)
        if v(api_version) >= v("2.2"):
            check(
                lambda: [n.replace("+", " ") for n in client.torrents_categories()],
                name,
                reverse=True,
            )
            save_path_key = _categories_save_path_key(api_version)
            check(
                lambda: [
                    mkpath(cat[save_path_key])
                    for cat in client.torrents_categories().values()
                ],
                mkpath(save_path) or "",
                reverse=True,
            )
        if v(api_version) >= v("2.8.4") and enable_download_path is not False:
            check(
                lambda: [
                    mkpath(cat.get("download_path", ""))
                    for cat in client.torrents_categories().values()
                ],
                mkpath(download_path) or "",
                reverse=True,
            )
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_before_api_version("2.1.0")
@pytest.mark.parametrize(
    "edit_cat_func",
    [
        "torrents_edit_category",
        "torrents_editCategory",
        "torrent_categories.edit_category",
        "torrent_categories.editCategory",
    ],
)
@pytest.mark.parametrize("filepath", ["", "/tmp/"])
@pytest.mark.parametrize("name", ["editcategory"])
@pytest.mark.parametrize("enable_download_path", [None, True, False])
def test_edit_category(
    client, api_version, edit_cat_func, filepath, name, enable_download_path
):
    try:
        client.torrents_create_category(
            name=name, save_path="/tmp/savetmp", download_path="/tmp/savetmp"
        )
        save_path = mkpath(filepath + "save/")
        download_path = mkpath(filepath + "down/")
        client.func(edit_cat_func)(
            name=name,
            save_path=save_path,
            download_path=download_path,
            enable_download_path=enable_download_path,
        )
        check(
            lambda: [n.replace("+", " ") for n in client.torrents_categories()],
            name,
            reverse=True,
        )
        save_path_key = _categories_save_path_key(api_version)
        check(
            lambda: (
                mkpath(cat[save_path_key])
                for cat in client.torrents_categories().values()
            ),
            mkpath(save_path) or "",
            reverse=True,
        )
        if v(api_version) >= v("2.8.4") and enable_download_path is not False:
            check(
                lambda: [
                    mkpath(cat.get("download_path", ""))
                    for cat in client.torrents_categories().values()
                ],
                mkpath(download_path) or "",
                reverse=True,
            )
    finally:
        client.torrents_remove_categories(categories=name)


@pytest.mark.skipif_after_api_version("2.1.0")
@pytest.mark.parametrize(
    "edit_cat_func",
    (
        "torrents_edit_category",
        "torrents_editCategory",
        "torrent_categories.edit_category",
        "torrent_categories.editCategory",
    ),
)
def test_edit_category_not_implemented(client, edit_cat_func):
    with pytest.raises(NotImplementedError):
        client.func(edit_cat_func)()


@pytest.mark.parametrize(
    "remove_cat_func",
    [
        "torrents_remove_categories",
        "torrents_removeCategories",
        "torrent_categories.remove_categories",
        "torrent_categories.removeCategories",
    ],
)
@pytest.mark.parametrize("categories", [["category1"], ["category1", "category 2"]])
def test_remove_category(
    client, api_version, orig_torrent, remove_cat_func, categories
):
    for name in categories:
        client.torrents_create_category(name=name)
    orig_torrent.set_category(category=categories[0])
    client.func(remove_cat_func)(categories=categories)
    if v(api_version) >= v("2.2"):
        check(
            lambda: [n.replace("+", " ") for n in client.torrents_categories()],
            categories,
            reverse=True,
            negate=True,
        )
    check(lambda: orig_torrent.info.category, categories[0], negate=True)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "tags_func",
    [
        "torrents_tags",
        "torrent_tags.tags",
    ],
)
def test_tags(client, tags_func):
    try:
        assert isinstance(client.func(tags_func)(), TagList)
    except TypeError:
        assert isinstance(client.func(tags_func), TagList)


@pytest.mark.skipif_before_api_version("2.3.0")
def test_tags_slice(client):
    assert isinstance(client.torrents_tags()[1:2], TagList)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "tags_func",
    [
        "torrents_tags",
        "torrent_tags.tags",
    ],
)
def test_tags_not_implemented(client, tags_func):
    with pytest.raises(NotImplementedError):
        client.func(tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
def test_add_tag_though_property(client):
    name = "newtag"
    client.torrent_tags.tags = name
    assert name in client.torrent_tags.tags
    client.torrent_tags.delete_tags(name)
    assert name not in client.torrent_tags.tags


@pytest.mark.skipif_after_api_version("2.3.0")
def test_add_tag_though_property_not_implemented(client):
    with pytest.raises(NotImplementedError):
        client.torrent_tags.tags = "asdf"


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func",
    [
        "torrents_add_tags",
        "torrents_addTags",
        "torrent_tags.add_tags",
        "torrent_tags.addTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_add_tags(client, orig_torrent, add_tags_func, tags):
    try:
        client.func(add_tags_func)(tags=tags, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.tags, tags, reverse=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func",
    [
        "torrents_add_tags",
        "torrents_addTags",
        "torrent_tags.add_tags",
        "torrent_tags.addTags",
    ],
)
def test_add_tags_not_implemented(client, add_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(add_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "remove_tags_func",
    [
        "torrents_remove_tags",
        "torrents_removeTags",
        "torrent_tags.remove_tags",
        "torrent_tags.removeTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_remove_tags(client, orig_torrent, remove_tags_func, tags):
    try:
        orig_torrent.add_tags(tags=tags)
        client.func(remove_tags_func)(tags=tags, torrent_hashes=orig_torrent.hash)
        check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "remove_tags_func",
    [
        "torrents_remove_tags",
        "torrents_removeTags",
        "torrent_tags.remove_tags",
        "torrent_tags.removeTags",
    ],
)
def test_remove_tags_not_implemented(client, remove_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(remove_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "create_tags_func",
    [
        "torrents_create_tags",
        "torrents_createTags",
        "torrent_tags.create_tags",
        "torrent_tags.createTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_create_tags(client, create_tags_func, tags):
    try:
        client.func(create_tags_func)(tags=tags)
        check(lambda: client.torrents_tags(), tags, reverse=True)
    finally:
        client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "create_tags_func",
    [
        "torrents_create_tags",
        "torrents_createTags",
        "torrent_tags.create_tags",
        "torrent_tags.createTags",
    ],
)
def test_create_tags_not_implemented(client, create_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(create_tags_func)()


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "delete_tags_func",
    [
        "torrents_delete_tags",
        "torrents_deleteTags",
        "torrent_tags.delete_tags",
        "torrent_tags.deleteTags",
    ],
)
@pytest.mark.parametrize("tags", [["tag1"], ["tag1", "tag 2"]])
def test_delete_tags(client, delete_tags_func, tags):
    client.torrents_create_tags(tags=tags)
    client.func(delete_tags_func)(tags=tags)
    check(lambda: client.torrents_tags(), tags, reverse=True, negate=True)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "delete_tags_func",
    [
        "torrents_delete_tags",
        "torrents_deleteTags",
        "torrent_tags.delete_tags",
        "torrent_tags.deleteTags",
    ],
)
def test_delete_tags_not_implemented(client, delete_tags_func):
    with pytest.raises(NotImplementedError):
        client.func(delete_tags_func)()



import sys

import pytest

from qbittorrentapi import APINames
from qbittorrentapi import Client
from qbittorrentapi.exceptions import APIConnectionError


@pytest.mark.skipif(sys.version_info < (3, 9), reason="removeprefix not in 3.8")
def test_methods(client):
    namespace = APINames.Authorization
    all_dotted_methods = set(dir(getattr(client, namespace)))

    for meth in [meth for meth in dir(client) if meth.startswith(f"{namespace}_")]:
        assert meth.removeprefix(f"{namespace}_") in all_dotted_methods


def test_is_logged_in():
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client.is_logged_in is False

    client.auth_log_in()
    assert client.is_logged_in is True

    client.auth_log_out()
    assert client.is_logged_in is False


def test_is_logged_in_bad_client():
    client = Client(
        host="asdf",
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client.is_logged_in is False

    with pytest.raises(APIConnectionError):
        client.auth_log_in()
    assert client.is_logged_in is False

    with pytest.raises(APIConnectionError):
        client.auth_log_out()
    assert client.is_logged_in is False


def test_session_cookie(app_version):
    client = Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    )
    assert client._session_cookie() is None

    # make the client perform a login
    assert client.app.version == app_version

    # should test other cookie names but it's difficult to change
    curr_sess_cookie = client._http_session.cookies["SID"]
    assert client._SID == curr_sess_cookie
    assert client._session_cookie() == curr_sess_cookie


def test_login_context_manager():
    with Client(
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=True,
        VERIFY_WEBUI_CERTIFICATE=False,
    ) as client:
        assert client.is_logged_in
    assert not client.is_logged_in



import platform
from time import sleep
from types import MethodType
from unittest.mock import MagicMock

import pytest

from qbittorrentapi import APINames
from qbittorrentapi import Conflict409Error
from qbittorrentapi import TorrentDictionary
from qbittorrentapi import TorrentFilesList
from qbittorrentapi import TorrentPieceInfoList
from qbittorrentapi import TorrentPropertiesDictionary
from qbittorrentapi import TorrentStates
from qbittorrentapi import TrackersList
from qbittorrentapi import WebSeedsList
from qbittorrentapi._version_support import v
from tests.test_torrents import disable_queueing
from tests.test_torrents import enable_queueing
from tests.utils import check
from tests.utils import mkpath
from tests.utils import retry


def test_info(orig_torrent, monkeypatch):
    assert orig_torrent.info.hash == orig_torrent.hash
    # mimic <=v2.0.1 where torrents_info() doesn't support hash arg
    with monkeypatch.context() as m:
        m.setattr(orig_torrent._client, "app_version", MagicMock(return_value="2.0.0"))
        assert orig_torrent.info.hash == orig_torrent.hash

    # ensure if things are really broken, an empty TorrentDictionary is returned...
    if platform.python_implementation() == "CPython":
        with monkeypatch.context() as m:

            def fake_info(*args, **kwargs):
                return []

            m.setattr(orig_torrent._client, "torrents_info", fake_info)
            assert isinstance(orig_torrent.info, TorrentDictionary)
            assert bool(orig_torrent.info) is False


def test_sync_local(orig_torrent):
    orig_torrent.state = "gibberish"
    assert orig_torrent.state == "gibberish"
    orig_torrent.sync_local()
    assert isinstance(orig_torrent, TorrentDictionary)
    assert orig_torrent.state != "gibberish"


def test_state_enum(orig_torrent):
    assert orig_torrent.state_enum in TorrentStates
    assert orig_torrent.state_enum is not TorrentStates.UNKNOWN
    orig_torrent.resume()
    check(
        lambda: orig_torrent.sync_local() is None
        and orig_torrent.state_enum.is_downloading,
        True,
    )
    # simulate an unknown torrent.state
    orig_torrent.state = "gibberish"
    assert orig_torrent.state_enum is TorrentStates.UNKNOWN
    # restore torrent state
    orig_torrent.sync_local()
    check(lambda: orig_torrent.state_enum.is_downloading, True)


# test fails on 4.1.0 release
@pytest.mark.skipif_before_api_version("2.0.1")
def test_pause_resume(client, new_torrent):
    new_torrent.pause()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        True,
    )

    new_torrent.resume()
    check(
        lambda: client.torrents_info(torrent_hashes=new_torrent.hash)[
            0
        ].state_enum.is_paused,
        False,
    )


@pytest.mark.parametrize("delete", [True, False, None, 0, 1])
def test_delete(client_mock, new_torrent, delete):
    new_torrent.delete(delete_files=delete)
    client_mock._post.assert_called_with(
        _name=APINames.Torrents,
        _method="delete",
        data={
            "hashes": new_torrent.hash,
            "deleteFiles": bool(delete),
        },
    )
    check(
        lambda: [t.hash for t in client_mock.torrents_info()],
        new_torrent.hash,
        reverse=True,
        negate=True,
    )


@pytest.mark.parametrize(
    "inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func",
    [
        ("increase_priority", "decrease_priority", "top_priority", "bottom_priority"),
        ("increasePrio", "decreasePrio", "topPrio", "bottomPrio"),
    ],
)
def test_priority(
    client, new_torrent, inc_prio_func, dec_prio_func, top_prio_func, bottom_prio_func
):
    disable_queueing(client)

    with pytest.raises(Conflict409Error):
        new_torrent.func(inc_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(dec_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(top_prio_func)()
    with pytest.raises(Conflict409Error):
        new_torrent.func(bottom_prio_func)()

    enable_queueing(client)
    sleep(0.25)  # putting sleeps in since these keep crashing qbittorrent

    current_priority = new_torrent.info.priority
    new_torrent.func(inc_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(dec_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority > current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(top_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority < current_priority, True)

    current_priority = new_torrent.info.priority
    new_torrent.func(bottom_prio_func)()
    sleep(0.25)
    check(lambda: new_torrent.info.priority > current_priority, True)


@pytest.mark.skipif_before_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func", ["set_share_limits", "setShareLimits"]
)
def test_set_share_limits(orig_torrent, set_share_limits_func):
    orig_torrent.func(set_share_limits_func)(
        ratio_limit=5, seeding_time_limit=100, inactive_seeding_time_limit=200
    )
    check(lambda: orig_torrent.info.max_ratio, 5)
    check(lambda: orig_torrent.info.max_seeding_time, 100)
    if "max_inactive_seeding_time" in orig_torrent.info:
        check(lambda: orig_torrent.info.max_inactive_seeding_time, 200)


@pytest.mark.skipif_after_api_version("2.0.1")
@pytest.mark.parametrize(
    "set_share_limits_func", ["set_share_limits", "setShareLimits"]
)
def test_set_share_limits_not_implemented(orig_torrent, set_share_limits_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(set_share_limits_func)(ratio_limit=5, seeding_time_limit=100)


@pytest.mark.parametrize(
    "down_limit_func, set_down_limit_func",
    [("download_limit", "set_download_limit"), ("downloadLimit", "setDownloadLimit")],
)
def test_download_limit(orig_torrent, down_limit_func, set_down_limit_func):
    setattr(orig_torrent, down_limit_func, 2048)
    check(lambda: orig_torrent.func(down_limit_func), 2048)
    check(lambda: orig_torrent.info.dl_limit, 2048)

    orig_torrent.func(set_down_limit_func)(4096)
    check(lambda: orig_torrent.func(down_limit_func), 4096)
    check(lambda: orig_torrent.info.dl_limit, 4096)


@pytest.mark.parametrize(
    "up_limit_func, set_up_limit_func",
    [("upload_limit", "set_upload_limit"), ("uploadLimit", "setUploadLimit")],
)
def test_upload_limit(orig_torrent, up_limit_func, set_up_limit_func):
    setattr(orig_torrent, up_limit_func, 2048)
    check(lambda: orig_torrent.func(up_limit_func), 2048)
    check(lambda: orig_torrent.info.up_limit, 2048)

    orig_torrent.func(set_up_limit_func)(4096)
    check(lambda: orig_torrent.func(up_limit_func), 4096)
    check(lambda: orig_torrent.info.up_limit, 4096)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize("set_loc_func", ["set_location", "setLocation"])
def test_set_location(new_torrent, set_loc_func, tmp_path):
    sleep(0.5)
    loc = mkpath(tmp_path, "3")
    new_torrent.func(set_loc_func)(loc)
    check(lambda: mkpath(new_torrent.info.save_path), mkpath(loc))


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("set_save_path_func", ["set_save_path", "setSavePath"])
def test_set_save_path(new_torrent, set_save_path_func, tmp_path):
    loc = mkpath(tmp_path, "savepath3")
    new_torrent.func(set_save_path_func)(loc)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.save_path), mkpath(loc))


@pytest.mark.skipif_before_api_version("2.8.4")
@pytest.mark.parametrize("set_down_path_func", ["set_download_path", "setDownloadPath"])
def test_set_download_path(new_torrent, set_down_path_func, tmp_path):
    loc = mkpath(tmp_path, "downloadpath3")
    new_torrent.func(set_down_path_func)(loc)
    # qBittorrent may return trailing separators depending on version....
    check(lambda: mkpath(new_torrent.info.download_path), mkpath(loc))


@pytest.mark.parametrize("set_cat_func", ["set_category", "setCategory"])
@pytest.mark.parametrize("category", ["category 1", "category_1"])
def test_set_category(client, orig_torrent, set_cat_func, category):
    client.torrents_create_category(category=category)
    orig_torrent.func(set_cat_func)(category=category)
    check(lambda: orig_torrent.info.category.replace("+", " "), category, reverse=True)
    client.torrents_remove_categories(categories=category)


@pytest.mark.parametrize(
    "set_auto_mgmt_func", ["set_auto_management", "setAutoManagement"]
)
def test_set_auto_management(orig_torrent, set_auto_mgmt_func):
    current_setting = orig_torrent.auto_tmm
    orig_torrent.func(set_auto_mgmt_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.auto_tmm, not current_setting)
    orig_torrent.func(set_auto_mgmt_func)(enable=current_setting)
    check(lambda: orig_torrent.info.auto_tmm, current_setting)


@pytest.mark.parametrize(
    "toggle_seq_down_func", ["toggle_sequential_download", "toggleSequentialDownload"]
)
def test_toggle_sequential_download(orig_torrent, toggle_seq_down_func):
    current_setting = orig_torrent.seq_dl
    orig_torrent.func(toggle_seq_down_func)()
    check(lambda: orig_torrent.info.seq_dl, not current_setting)
    orig_torrent.func(toggle_seq_down_func)()
    check(lambda: orig_torrent.info.seq_dl, current_setting)


@pytest.mark.skipif_before_api_version("2.0.2")
@pytest.mark.parametrize(
    "toggle_piece_prio_func",
    ["toggle_first_last_piece_priority", "toggleFirstLastPiecePrio"],
)
def test_toggle_first_last_piece_priority(orig_torrent, toggle_piece_prio_func):
    current_setting = orig_torrent.f_l_piece_prio
    orig_torrent.func(toggle_piece_prio_func)()
    check(lambda: orig_torrent.info.f_l_piece_prio, not current_setting)


@pytest.mark.parametrize("set_force_start_func", ["set_force_start", "setForceStart"])
def test_set_force_start(orig_torrent, set_force_start_func):
    current_setting = orig_torrent.force_start
    orig_torrent.func(set_force_start_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.force_start, not current_setting)
    orig_torrent.func(set_force_start_func)(enable=current_setting)
    check(lambda: orig_torrent.info.force_start, current_setting)


@pytest.mark.parametrize(
    "set_super_seeding_func", ["set_super_seeding", "setSuperSeeding"]
)
def test_set_super_seeding(orig_torrent, set_super_seeding_func):
    current_setting = orig_torrent.super_seeding
    orig_torrent.func(set_super_seeding_func)(enable=(not current_setting))
    check(lambda: orig_torrent.info.super_seeding, not current_setting)
    orig_torrent.func(set_super_seeding_func)(enable=current_setting)
    check(lambda: orig_torrent.info.super_seeding, current_setting)


def test_properties(orig_torrent):
    assert isinstance(orig_torrent.properties, TorrentPropertiesDictionary)
    assert "save_path" in orig_torrent.properties


@pytest.mark.parametrize("trackers", ["127.0.0.2", ["127.0.0.3", "127.0.0.4"]])
def test_trackers(orig_torrent, trackers):
    assert isinstance(orig_torrent.trackers, TrackersList)
    assert "num_peers" in orig_torrent.trackers[-1]

    orig_torrent.trackers = trackers
    check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)


@pytest.mark.parametrize("add_trackers_func", ["add_trackers", "addTrackers"])
@pytest.mark.parametrize("trackers", ["127.0.0.2", ["127.0.0.3", "127.0.0.4"]])
def test_add_tracker(new_torrent, add_trackers_func, trackers):
    new_torrent.func(add_trackers_func)(urls=trackers)
    sleep(0.1)  # try to stop crashing qbittorrent
    check(lambda: (t.url for t in new_torrent.trackers), trackers, reverse=True)


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize("edit_tracker_func", ["edit_tracker", "editTracker"])
def test_edit_tracker(orig_torrent, edit_tracker_func):
    orig_torrent.add_trackers(urls="127.0.1.1")
    orig_torrent.func(edit_tracker_func)(orig_url="127.0.1.1", new_url="127.0.1.2")
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        "127.0.1.1",
        reverse=True,
        negate=True,
    )
    check(lambda: (t.url for t in orig_torrent.trackers), "127.0.1.2", reverse=True)
    orig_torrent.remove_trackers(urls="127.0.1.2")


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize("edit_tracker_func", ["edit_tracker", "editTracker"])
def test_edit_tracker_not_implemented(orig_torrent, edit_tracker_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(edit_tracker_func)()


@pytest.mark.skipif_before_api_version("2.2.0")
@pytest.mark.parametrize("remove_trackers_func", ["remove_trackers", "removeTrackers"])
@pytest.mark.parametrize("trackers", ["127.0.2.2", ["127.0.2.3", "127.0.2.4"]])
def test_remove_trackers(orig_torrent, remove_trackers_func, trackers):
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )
    orig_torrent.add_trackers(urls=trackers)
    check(lambda: (t.url for t in orig_torrent.trackers), trackers, reverse=True)
    orig_torrent.func(remove_trackers_func)(urls=trackers)
    check(
        lambda: (t.url for t in orig_torrent.trackers),
        trackers,
        reverse=True,
        negate=True,
    )


@pytest.mark.skipif_after_api_version("2.2.0")
@pytest.mark.parametrize("remove_trackers_func", ["remove_trackers", "removeTrackers"])
def test_remove_trackers_not_implemented(orig_torrent, remove_trackers_func):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(remove_trackers_func)()


def test_webseeds(orig_torrent):
    assert isinstance(orig_torrent.webseeds, WebSeedsList)


def test_files(orig_torrent):
    assert isinstance(orig_torrent.files, TorrentFilesList)
    assert "id" in orig_torrent.files[0]


def test_recheck(orig_torrent):
    orig_torrent.recheck()


@pytest.mark.skipif_before_api_version("2.0.2")
def test_reannounce(orig_torrent):
    orig_torrent.reannounce()


@pytest.mark.skipif_after_api_version("2.0.2")
def test_reannounce_not_implemented(orig_torrent):
    with pytest.raises(NotImplementedError):
        orig_torrent.reannounce()


@pytest.mark.skipif_before_api_version("2.10")
def test_reannounce_in(orig_torrent):
    assert "reannounce_in" in orig_torrent.info
    assert type(orig_torrent.reannounce) is MethodType


@pytest.mark.skipif_after_api_version("2.10")
def test_reannounce_in_not_present(orig_torrent):
    assert "reannounce_in" not in orig_torrent.info


@pytest.mark.skipif_before_api_version("2.4.0")
@pytest.mark.parametrize("rename_file_func", ["rename_file", "renameFile"])
@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename_file(app_version, new_torrent, rename_file_func, name):
    @retry()
    def run_test_old():
        new_torrent.func(rename_file_func)(file_id=0, new_file_name=name)
        check(lambda: new_torrent.files[0].name, name)

    run_test_old()

    if v(app_version) >= v("v4.3.3"):

        @retry()
        def run_test_new():
            curr_name = new_torrent.files[0].name
            new_name = "NEW_" + name
            new_torrent.func(rename_file_func)(old_path=curr_name, new_path=new_name)
            check(lambda: new_torrent.files[0].name, new_name)

        run_test_new()


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize("rename_file_func", ["rename_file", "renameFile"])
def test_rename_file_not_implemented(new_torrent, rename_file_func):
    with pytest.raises(NotImplementedError):
        new_torrent.func(rename_file_func)()


@pytest.mark.skipif_before_api_version("2.7")
@pytest.mark.parametrize("rename_folder_func", ["rename_folder", "renameFolder"])
@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename_folder(app_version, new_torrent, rename_folder_func, name):
    # need to ensure we're at least on v4.3.3 to run test since
    # both v4.3.2 and v4.3.2 both use Web API 2.7
    if v(app_version) >= v("v4.3.3"):
        # move the file in to a new folder
        orig_file_path = new_torrent.files[0].name
        new_folder = "qwer"
        new_torrent.rename_file(
            old_path=orig_file_path,
            new_path=new_folder + "/" + orig_file_path,
        )

        # wait for the folder to be renamed
        check(
            lambda: [f.name.split("/")[0] for f in new_torrent.files],
            new_folder,
            reverse=True,
        )

        # test rename that new folder
        new_torrent.func(rename_folder_func)(
            old_path=new_folder,
            new_path=name,
        )
        check(
            lambda: new_torrent.files[0].name.replace("+", " "),
            name + "/" + orig_file_path,
        )


@pytest.mark.skipif_after_api_version("2.4.0")
@pytest.mark.parametrize("rename_folder_func", ["rename_folder", "renameFolder"])
def test_rename_folder_not_implemented(new_torrent, rename_folder_func):
    with pytest.raises(NotImplementedError):
        new_torrent.func(rename_folder_func)()


@pytest.mark.skipif_before_api_version("2.8.14")
def test_export(orig_torrent):
    assert isinstance(orig_torrent.export(), bytes)


@pytest.mark.skipif_after_api_version("2.8.14")
def test_export_not_implemented(orig_torrent):
    with pytest.raises(NotImplementedError):
        orig_torrent.export()


@pytest.mark.parametrize("piece_states_func", ["piece_states", "pieceStates"])
def test_piece_states(orig_torrent, piece_states_func):
    assert isinstance(orig_torrent.func(piece_states_func), TorrentPieceInfoList)


@pytest.mark.parametrize("piece_hashes_func", ["piece_hashes", "pieceHashes"])
def test_piece_hashes(orig_torrent, piece_hashes_func):
    assert isinstance(orig_torrent.func(piece_hashes_func), TorrentPieceInfoList)


@pytest.mark.parametrize("file_prio_func", ["file_priority", "filePriority"])
def test_file_priority(orig_torrent, file_prio_func):
    orig_torrent.func(file_prio_func)(file_ids=0, priority=7)
    check(lambda: orig_torrent.files[0].priority, 7)


@pytest.mark.parametrize("name", ["new_name", "new name"])
def test_rename(new_torrent, name):
    new_torrent.rename(new_name=name)
    check(lambda: new_torrent.info.name.replace("+", " "), name)


@pytest.mark.skipif_before_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func, remove_tags_func",
    [("add_tags", "remove_tags"), ("addTags", "removeTags")],
)
@pytest.mark.parametrize("tags", ["tag 1", ["tag 2", "tag 3"]])
def test_add_remove_tags(client, orig_torrent, add_tags_func, remove_tags_func, tags):
    orig_torrent.func(add_tags_func)(tags=tags)
    check(lambda: orig_torrent.info.tags, tags, reverse=True)

    orig_torrent.func(remove_tags_func)(tags=tags)
    check(lambda: orig_torrent.info.tags, tags, reverse=True, negate=True)

    client.torrents_delete_tags(tags=tags)


@pytest.mark.skipif_after_api_version("2.3.0")
@pytest.mark.parametrize(
    "add_tags_func, remove_tags_func",
    [("add_tags", "remove_tags"), ("addTags", "removeTags")],
)
def test_add_remove_tags_not_implemented(
    client, orig_torrent, add_tags_func, remove_tags_func
):
    with pytest.raises(NotImplementedError):
        orig_torrent.func(add_tags_func)()
    with pytest.raises(NotImplementedError):
        orig_torrent.func(remove_tags_func)()



from enum import Enum

import pytest

from qbittorrentapi._attrdict import AttrDict
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.definitions import TorrentState
from qbittorrentapi.definitions import TrackerStatus

torrent_all_states = [
    "error",
    "missingFiles",
    "uploading",
    "pausedUP",
    "queuedUP",
    "stalledUP",
    "checkingUP",
    "forcedUP",
    "allocating",
    "downloading",
    "metaDL",
    "forcedMetaDL",
    "pausedDL",
    "queuedDL",
    "forcedDL",
    "stalledDL",
    "checkingDL",
    "checkingResumeData",
    "moving",
    "unknown",
]

torrent_downloading_states = [
    "downloading",
    "metaDL",
    "forcedMetaDL",
    "stalledDL",
    "checkingDL",
    "pausedDL",
    "queuedDL",
    "forcedDL",
]

torrent_uploading_states = [
    "uploading",
    "stalledUP",
    "checkingUP",
    "queuedUP",
    "forcedUP",
]

torrent_complete_states = [
    "uploading",
    "stalledUP",
    "checkingUP",
    "pausedUP",
    "queuedUP",
    "forcedUP",
]

torrent_checking_states = ["checkingUP", "checkingDL", "checkingResumeData"]

torrent_errored_states = ["missingFiles", "error"]

torrent_paused_states = ["pausedUP", "pausedDL"]


def test_torrent_states_exists():
    assert isinstance(TorrentState("unknown"), Enum)


@pytest.mark.parametrize("state", torrent_all_states)
def test_all_states(state):
    assert TorrentState(state) in TorrentState


@pytest.mark.parametrize("state", torrent_all_states)
def test_states_cmp_str(state):
    assert TorrentState(state) == TorrentState(state).value


@pytest.mark.parametrize("state", torrent_downloading_states)
def test_downloading_states(state):
    assert TorrentState(state).is_downloading


@pytest.mark.parametrize("state", torrent_uploading_states)
def test_uploading_states(state):
    assert TorrentState(state).is_uploading


@pytest.mark.parametrize("state", torrent_complete_states)
def test_completing_states(state):
    assert TorrentState(state).is_complete


@pytest.mark.parametrize("state", torrent_checking_states)
def test_checking_states(state):
    assert TorrentState(state).is_checking


@pytest.mark.parametrize("state", torrent_errored_states)
def test_errored_states(state):
    assert TorrentState(state).is_errored


@pytest.mark.parametrize("state", torrent_paused_states)
def test_paused_states(state):
    assert TorrentState(state).is_paused


@pytest.mark.parametrize("status", [0, 1, 2, 3, 4])
def test_tracker_statuses_exist(status):
    assert isinstance(TrackerStatus(status), Enum)


@pytest.mark.parametrize("status", [0, 1, 2, 3, 4])
def test_tracker_statuses_cmp_int(status):
    assert TrackerStatus(status) == TrackerStatus(status).value


@pytest.mark.parametrize(
    "status, display_value",
    [
        (0, "Disabled"),
        (1, "Not contacted"),
        (2, "Working"),
        (3, "Updating"),
        (4, "Not working"),
    ],
)
def test_tracker_status_display(status, display_value):
    assert TrackerStatus(status).display == display_value


def test_testing_groups():
    assert set(torrent_all_states) == {s.value for s in iter(TorrentState)}
    assert set(torrent_downloading_states) == {
        s.value for s in TorrentState if s.is_downloading
    }
    assert set(torrent_uploading_states) == {
        s.value for s in TorrentState if s.is_uploading
    }
    assert set(torrent_complete_states) == {
        s.value for s in TorrentState if s.is_complete
    }
    assert set(torrent_checking_states) == {
        s.value for s in TorrentState if s.is_checking
    }
    assert set(torrent_errored_states) == {
        s.value for s in TorrentState if s.is_errored
    }


def test_dictionary():
    assert len(Dictionary()) == 0
    assert len(Dictionary({"one": {"two": 2}})) == 1
    assert isinstance(Dictionary({"one": {"two": 2}}).one, AttrDict)
    assert isinstance(Dictionary({"one": {"two": 2}})["one"], AttrDict)
    assert Dictionary({"one": {"two": 2}}).one.two == 2
    assert Dictionary({"one": {"two": 2}})["one"]["two"] == 2
    assert Dictionary({"three": 3}).three == 3
    assert Dictionary({"three": 3})["three"] == 3


def test_list(client):
    assert len(List()) == 0
    list_entries = [{"one": "1"}, {"two": "2"}, {"three": "3"}]
    test_list = List(list_entries, entry_class=ListEntry)
    assert len(test_list) == 3
    assert issubclass(type(test_list[0]), ListEntry)
    assert test_list[0].one == "1"


def test_list_actions(client):
    list_one = List(
        [{"one": "1"}, {"two": "2"}, {"three": "3"}],
        entry_class=ListEntry,
    )
    list_two = List([{"four": "4"}], entry_class=ListEntry)

    assert list_one[1:3] == [
        ListEntry({"two": "2"}),
        ListEntry({"three": "3"}),
    ]
    assert list_one + list_two == [
        ListEntry({"one": "1"}),
        ListEntry({"two": "2"}),
        ListEntry({"three": "3"}),
    ] + [ListEntry({"four": "4"})]

    assert list_one.copy() == [
        ListEntry({"one": "1"}),
        ListEntry({"two": "2"}),
        ListEntry({"three": "3"}),
    ]



from __future__ import annotations

from functools import wraps
from logging import Logger
from logging import getLogger
from types import TracebackType
from typing import TYPE_CHECKING

from requests import Response

from qbittorrentapi import Version
from qbittorrentapi.definitions import APIKwargsT
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.exceptions import LoginFailed
from qbittorrentapi.exceptions import UnsupportedQbittorrentVersion
from qbittorrentapi.request import Request

if TYPE_CHECKING:
    from qbittorrentapi.client import Client

logger: Logger = getLogger(__name__)


class AuthAPIMixIn(Request):
    """
    Implementation of all ``Authorization`` API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> _ = client.is_logged_in
        >>> client.auth_log_in(username="admin", password="adminadmin")
        >>> client.auth_log_out()
    """

    @property
    def auth(self) -> Authorization:
        """Allows for transparent interaction with Authorization endpoints."""
        if self._authorization is None:
            self._authorization = Authorization(client=self)
        return self._authorization

    authorization = auth

    @property
    def is_logged_in(self) -> bool:
        """
        Returns True if low-overhead API call succeeds; False otherwise.

        There isn't a reliable way to know if an existing session is still valid without
        attempting to use it. qBittorrent invalidates cookies when they expire.

        :returns: ``True``/``False`` if current authorization cookie is accepted by qBittorrent
        """
        try:
            # use _request_manager() directly so log in is not attempted
            self._request_manager(
                http_method="post",
                api_namespace=APINames.Application,
                api_method="version",
            )
        except Exception:
            return False
        else:
            return True

    def auth_log_in(
        self,
        username: str | None = None,
        password: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        """
        Log in to qBittorrent host.

        :raises LoginFailed: if credentials failed to log in
        :raises Forbidden403Error: if user is banned...or not logged in

        :param username: username for qBittorrent client
        :param password: password for qBittorrent client
        """
        if username:
            self.username = username
            self._password = password or ""

        # trigger a (re-)initialization in case this is a new instance of qBittorrent
        self._initialize_context()

        creds = {"username": self.username, "password": self._password}
        auth_response = self._post_cast(
            _name=APINames.Authorization,
            _method="login",
            data=creds,
            response_class=Response,
            **kwargs,
        )

        if auth_response.text != "Ok.":
            logger.debug("Login failed")
            raise LoginFailed()
        logger.debug("Login successful")

        # check if the connected qBittorrent is fully supported by this Client yet
        if self._RAISE_UNSUPPORTEDVERSIONERROR:
            app_version = self.app_version()  # type: ignore
            api_version = self.app_web_api_version()  # type: ignore
            if not (
                Version.is_api_version_supported(api_version)
                and Version.is_app_version_supported(app_version)
            ):
                raise UnsupportedQbittorrentVersion(
                    "This version of qBittorrent is not fully supported => "
                    f"App Version: {app_version} API Version: {api_version}"
                )

    @property
    def _SID(self) -> str | None:
        """
        Authorization session cookie from qBittorrent using default cookie name `SID`.

        Backwards compatible for :meth:`~AuthAPIMixIn._session_cookie`.
        """
        return self._session_cookie()

    def _session_cookie(self, cookie_name: str = "SID") -> str | None:
        """
        Authorization session cookie from qBittorrent.

        :param cookie_name: Name of the authorization cookie; configurable after v4.5.0.
        """
        if self._http_session:
            return self._http_session.cookies.get(cookie_name, None)  # type: ignore
        return None

    def auth_log_out(self, **kwargs: APIKwargsT) -> None:
        """End session with qBittorrent."""
        self._post(_name=APINames.Authorization, _method="logout", **kwargs)

    def __enter__(self) -> Client:
        self.auth_log_in()
        return self  # type: ignore[return-value]

    def __exit__(
        self,
        exctype: type[BaseException] | None,
        excinst: BaseException | None,
        exctb: TracebackType | None,
    ) -> None:
        self.auth_log_out()


class Authorization(ClientCache[AuthAPIMixIn]):
    """
    Allows interaction with the ``Authorization`` API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host="localhost:8080", username="admin", password="adminadmin")
        >>> is_logged_in = client.auth.is_logged_in
        >>> client.auth.log_in(username="admin", password="adminadmin")
        >>> client.auth.log_out()
    """

    @property
    @wraps(AuthAPIMixIn.is_logged_in)
    def is_logged_in(self) -> bool:
        return self._client.is_logged_in

    @wraps(AuthAPIMixIn.auth_log_in)
    def log_in(
        self,
        username: str | None = None,
        password: str | None = None,
        **kwargs: APIKwargsT,
    ) -> None:
        return self._client.auth_log_in(username=username, password=password, **kwargs)

    @wraps(AuthAPIMixIn.auth_log_out)
    def log_out(self, **kwargs: APIKwargsT) -> None:
        return self._client.auth_log_out(**kwargs)
